@@ -69,13 +69,16 @@ MAP = NodeMapper()
 class Hint(enum.Flag):
     """Linebreak hinting when we write out otherwise formatted lines.
 
-    The formatters provide these hints based on their surrounding context.
+    The formatters provide these hints based on their surrounding context.  On
+    occasion, hinting is also used to pass flags from higher-level (in the tree)
+    to lower-level Formatters.
     """
     NONE = enum.auto()
     GOOD_AFTER_LB = enum.auto() # A linebreak before this item is encouraged.
     NO_LB_BEFORE = enum.auto() # Never line-break before this item.
     NO_LB_AFTER = enum.auto() # Never line-break after this item.
     ZERO_WIDTH = enum.auto() # This item doesn't contribute to line length.
+    COMPLEX_BLOCK = enum.auto() # A {}-block is complex enough to linebreak
 
 
 class Formatter:
@@ -105,18 +108,19 @@ class Formatter:
         node.formatter = self
 
     def format(self):
+        """Default formatting for a tree node
+
+        This simply writes out children as per their own formatting, and writes
+        out tokens directly.
+        """
         if self.node.children:
             self._format_children()
         else:
             self._format_token()
 
-    def content(self):
-        """Returns the script content bytes this formatter processes."""
-        return self.script[self.node.start_byte:self.node.end_byte]
-
     def _next_child(self):
         try:
-            node = self.node.children[self._cidx]
+            node = self.node.nonerr_children[self._cidx]
             self._cidx += 1
             return node
         except IndexError:
@@ -129,17 +133,32 @@ class Formatter:
                            hints=hints)
         formatter.format()
 
-    def _format_child(self, indent=False, hints=None):
-        node = self._next_child()
+    def _format_child(self, child=None, indent=False, hints=None):
+        if child is None:
+            child = self._next_child()
 
-        for child in node.prev_cst_siblings:
-            self._format_child_impl(child, indent)
+        # XXX Pretty subtle that we handle the child's surrounding context here,
+        # in the parent. Might have to refactor in the future.
 
-        # The hints apply to AST (not CST) nodes
-        self._format_child_impl(node, indent, hints)
+        # If the node has any preceding errors, render these out first.  Do
+        # this via _format_child(), not _format_child_impl(), since the error
+        # nodes are full-blown AST nodes potentially with their own CST
+        # neighborhood.
+        for node in child.prev_error_siblings:
+            self._format_child(node, indent)
 
-        for child in node.next_cst_siblings:
-            self._format_child_impl(child, indent)
+        for node in child.prev_cst_siblings:
+            self._format_child_impl(node, indent)
+
+        # The hints apply to AST (not CST) nodes, so now:
+        self._format_child_impl(child, indent, hints)
+
+        for node in child.next_cst_siblings:
+            self._format_child_impl(node, indent)
+
+        # Mirroring the above, handle any trailing errors last.
+        for node in child.next_error_siblings:
+            self._format_child(node, indent)
 
     def _format_child_range(self, num, hints=None, first_hints=None):
         """Format a given number of children of the node.
@@ -156,20 +175,22 @@ class Formatter:
 
         if num <= 0:
             return
-        elif num == 1:
+
+        if num == 1:
             # Single element: general and first-element hinting
             self._format_child(hints=hints | first_hints)
-        else:
-            # First element of multiple: general hinting; first-element hinting;
-            # avoid line breaks after the element.
-            self._format_child(hints=hints | first_hints | Hint.NO_LB_AFTER)
+            return
 
-            # Inner elements: general hinting; avoid line breaks
-            for _ in range(num-2):
-                self._format_child(hints=hints | Hint.NO_LB_AFTER)
+        # First element of multiple: general hinting; first-element hinting;
+        # avoid line breaks after the element.
+        self._format_child(hints=hints | first_hints | Hint.NO_LB_AFTER) # pylint: disable=unsupported-binary-operation
 
-            # Last element: general hinting only.
-            self._format_child(hints=hints)
+        # Inner elements: general hinting; avoid line breaks
+        for _ in range(num-2):
+            self._format_child(hints=hints | Hint.NO_LB_AFTER) # pylint: disable=unsupported-binary-operation
+
+        # Last element: general hinting; avoid line break before
+        self._format_child(hints=hints | Hint.NO_LB_BEFORE) # pylint: disable=unsupported-binary-operation
 
     def _format_children(self, sep=None):
         """Format all children of the node.
@@ -188,9 +209,32 @@ class Formatter:
             self._format_child()
 
     def _format_token(self):
-        self._write(self.content())
+        self._write(self.script.get_content(*self.node.script_range()))
 
-    def _write(self, data):
+    def _format_curly_statement_list(self, indent=True):
+        """Format a child sequence of '{' <stmt_list>? '}'
+
+        The statement list is optional, and we need to take into account
+        comments to tweak the layout to "{ }" if there's really nothing between
+        the braces. The need to indent depends on the caller's context.
+        """
+        self._format_child(indent=indent, hints=Hint.NO_LB_BEFORE) # '{'
+
+        # Shorten braces to "{ }" if there is at most whitespace between them.
+        if (self._get_child_token() == '}' and
+            self._get_child().has_only_whitespace_before()):
+            self._write_sp()
+            self._format_child() # '}'
+            return
+
+        self._write_nl()
+
+        if self._get_child_type() == 'stmt_list':
+            self._format_child(indent=indent) # <stmt_list>
+            self._write_nl()
+        self._format_child(indent=indent) # '}'
+
+    def _write(self, data, raw=False):
         if isinstance(data, str):
             data = data.encode('UTF-8')
 
@@ -202,7 +246,7 @@ class Formatter:
             # would result without the presence of interrupting comments.
             data = data.lstrip()
 
-        self.ostream.write(data, self)
+        self.ostream.write(data, self, raw)
 
     def _write_indent(self):
         if self.ostream.get_column() == 0:
@@ -231,7 +275,7 @@ class Formatter:
 
     def _children_remaining(self):
         """Returns number of children of this node not yet visited."""
-        return len(self.node.children[self._cidx:])
+        return len(self.node.nonerr_children[self._cidx:])
 
     def _get_child(self, offset=0, absolute=False):
         """Accessor for child nodes, without adjusting the offset index.
@@ -247,7 +291,7 @@ class Formatter:
         cidx = 0 if absolute else self._cidx
 
         try:
-            return self.node.children[cidx + offset]
+            return self.node.nonerr_children[cidx + offset]
         except IndexError:
             return None
 
@@ -295,9 +339,9 @@ class Formatter:
     @staticmethod
     def lookup(node):
         """Formatter lookup for a zeekscript.Node, based on its type information."""
-        # If we're looking up a token node, always use a dummy formatter.
-        # This ensures that we don't confuse a node.type of the same name,
-        # e.g. a variable called 'decl'.
+        # If we're looking up a token node, always use the fallback formatter,
+        # which writes it out directly.  This ensures that we don't confuse a
+        # node.type of the same name, e.g. a variable called 'decl'.
         if not node.is_named:
             return Formatter
         return MAP.get(node.type)
@@ -309,6 +353,63 @@ class NullFormatter(Formatter):
         pass
 
 
+class ErrorFormatter(Formatter):
+    """This formatter handles parser errors reported by TreeSitter.
+
+    This is pretty tricky. Since we cannot really know what triggered the error,
+    this formatter preserves the node's byte range unmodified, expect for any
+    ranges covered by child nodes with their own nontrivial formatting (which
+    currently means child nodes that have children). This suits the behavior of
+    error nodes, which often still have functional child nodes, but surround
+    them with broken content.
+
+    If we rendered error nodes more simply, we'd either miss out on possible
+    formatting, introduce formatting that makes unpleasing changes (simple
+    space-separation will often look wrong, for example), or accidentally run
+    content together that the parser needs to remain whitespace-separated to
+    parse correctly.
+
+    To avoid merging the output directly with surrounding content (potentially
+    breaking subsequent script-parsing), this formatter adds a surrounding space
+    before and after the output. This could be optimized later via the notion of
+    an optional space that gets ignored when neighbored by other whitespace.
+    """
+    def format(self):
+        if not self.node.children:
+            content = self.script.get_content(*self.node.script_range())
+            self._write_sp()
+            self._write(content, raw=True)
+            self._write_sp()
+            return
+
+        start, _ = self.node.script_range()
+        _, end = self.node.script_range(with_cst=True)
+
+        # Before and after the error node's script range there may be
+        # non-newline whitespace skipped by the parser. It's tempting to tuck
+        # that onto the output, to preserve it. But, since formatting might
+        # itself tuck on whitespace, this can lead to repeated formatting
+        # continuously adding whitespace. So we ignore such space.
+
+        self._write_sp()
+
+        for node in self.node.children:
+            node_start, node_end = node.script_range()
+            if node.children:
+                if node_start > start:
+                    content = self.script.get_content(start, node_start)
+                    self._write(content, raw=True)
+
+                self._format_child(child=node)
+                start = node_end
+
+        if start < end:
+            content = self.script.get_content(start, end)
+            self._write(content, raw=True)
+
+        self._write_sp()
+
+
 class LineFormatter(Formatter):
     """This formatter separates all nodes with space and terminates with a newline."""
     def format(self):
@@ -348,6 +449,7 @@ class ModuleDeclFormatter(Formatter):
 
 class ExportDeclFormatter(Formatter):
     def format(self):
+        # No Whitesmith here: "{" on same line, closing "}" unindented.
         self._format_child() # 'export'
         self._write_sp()
         self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
@@ -390,6 +492,46 @@ class GlobalDeclFormatter(TypedInitializerFormatter):
         self._write_nl()
 
 
+class ComplexSequenceFormatterMixin():
+    """A mixin to figure out whether to line-break the (remaining) children.
+
+    The idea here is to determine if any one child is "complex". If so, the
+    entire sequence is to be written in individual lines. The formatting of
+    these lines doesn't happen in this mixin, only the complexity decision.
+
+    The default complexity decision simply looks for comments in the subtree.
+    """
+    def is_complex(self):
+        return self.is_complex_node(self.node)
+
+    def is_complex_node(self, node):
+        for node, _ in node.traverse(include_cst=True):
+            if node.is_comment():
+                return True
+        return False
+
+
+class ComplexBlockFormatterMixin(ComplexSequenceFormatterMixin):
+    """A mixin to figure out when a {}-block is "complex" enough to line-break it."""
+    def is_complex(self):
+        # The heuristic here is as in the parent class, but we don't operate on
+        # this formatter's whole tree. Instead, focus on the upcoming sequence
+        # of children, bounded by '{' and '}' tokens, since such sequences can
+        # be part of the node's larger sequence of children.
+        if not self._get_child() or self._get_child_token() != '{':
+            return False
+
+        offset = 0
+        while self._get_child(offset=offset):
+            if self.is_complex_node(self._get_child(offset=offset)):
+                return True
+            if self._get_child_token(offset=offset) == '}':
+                return False
+            offset += 1
+
+        return False
+
+
 class InitializerFormatter(Formatter):
     def format(self):
         if self._get_child_name() == 'init_class':
@@ -398,26 +540,81 @@ class InitializerFormatter(Formatter):
 
         self._format_child() # <init>
 
-class InitFormatter(Formatter):
+
+class InitFormatter(Formatter, ComplexBlockFormatterMixin):
+    """Initializers expand on the block "complexity" detection: in addition to the
+    newline logic, non-atomic expressions (things other than IDs and constants)
+    trigger line-breaking.
+    """
+    def is_complex_node(self, node):
+        if super().is_complex_node(node):
+            return True
+
+        # We only consider expressions here, so declare anything else simple.
+        if node.name() != 'expr':
+            return False
+
+        # Only "atomic" expressions (IDs, constants -- including patterns) count
+        # as simple.
+        return not (len(node.nonerr_children) == 1 and
+                    node.nonerr_children[0].name() in ('id', 'constant', 'pattern'))
+
     def format(self):
         if self._get_child_token() == '{':
+            # Any number of expressions, comma-separated, with optional final
+            # comma. We use the same heuristic as for enums: by default we keep
+            # elements on a single line, but in the presence of comments we
+            # break each expr onto a new line.
+            do_linebreak = self.is_complex() # Must call before we consume '{'
             self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
-            # Any number of expressions, comma-separated
+
             if self._get_child_name() == 'expr':
-                self._write_nl()
-                while self._get_child_name() == 'expr':
-                    self._format_child(indent=True) # <expr>
-                    if self._get_child_token() == ',':
-                        self._format_child(hints=Hint.NO_LB_BEFORE) # ','
+                if do_linebreak:
                     self._write_nl()
+                    while self._get_child_name() == 'expr':
+                        self._format_child(indent=True) # <expr>
+                        if self._get_child_token() == ',':
+                            self._format_child(hints=Hint.NO_LB_BEFORE) # ','
+                        self._write_nl()
+                else:
+                    self._write_sp()
+                    while self._get_child_name() == 'expr':
+                        self._format_child(indent=True) # <expr>
+                        if self._get_child_token() == ',':
+                            self._format_child(hints=Hint.NO_LB_BEFORE) # ','
+                        if self._get_child_name() == 'expr':
+                            self._write_sp()
+                    self._write_sp()
             else:
+                # Just a space when the initializer list has no members.
                 self._write_sp()
+
             self._format_child() # '}'
         else:
             self._format_child() # <expr>
 
 
-class RedefEnumDeclFormatter(Formatter):
+class EnumBodyFormatterMixin(ComplexBlockFormatterMixin):
+    """A mixin that knows when to break an enum_body onto lines."""
+    def _format_curly_enum_body(self):
+        """Formats an '{' <enum_body> '}' sequence."""
+        do_linebreak = self.is_complex() # Must call before we consume '{'
+
+        self._format_child() # '{'
+
+        if do_linebreak:
+            self._write_nl()
+            self._format_child(indent=True, hints=Hint.COMPLEX_BLOCK) # enum_body
+            self._write_nl()
+        else:
+            self._write_sp()
+            self._format_child(indent=True) # enum_body
+            self._write_sp()
+
+        self._format_child() # '}'
+
+
+class RedefEnumDeclFormatter(Formatter, EnumBodyFormatterMixin):
     def format(self):
         self._format_child() # 'redef'
         self._write_sp()
@@ -427,10 +624,8 @@ class RedefEnumDeclFormatter(Formatter):
         self._write_sp()
         self._format_child() # '+='
         self._write_sp()
-        self._format_child() # '{'
-        self._write_nl()
-        self._format_child(indent=True) # enum_body
-        self._format_child_range(2) # '}' ';'
+        self._format_curly_enum_body()
+        self._format_child(hints=Hint.NO_LB_BEFORE) # ';'
         self._write_nl()
 
 
@@ -470,7 +665,7 @@ class TypeDeclFormatter(Formatter):
         self._write_nl()
 
 
-class TypeFormatter(SpaceSeparatedFormatter):
+class TypeFormatter(SpaceSeparatedFormatter, EnumBodyFormatterMixin):
     def format(self):
         if self._get_child_token() == 'set':
             self._format_child() # 'set'
@@ -485,6 +680,7 @@ class TypeFormatter(SpaceSeparatedFormatter):
             self._format_child() # <type>
 
         elif self._get_child_token() == 'record':
+            # No Whitesmith here: "{" on same line, closing "}" unindented.
             self._format_child() # 'record',
             self._write_sp()
             self._format_child() # '{'
@@ -499,12 +695,10 @@ class TypeFormatter(SpaceSeparatedFormatter):
             self._format_child() # '}'
 
         elif self._get_child_token() == 'enum':
+            # No Whitesmith here: "{" on same line, closing "}" unindented.
             self._format_child() # 'enum'
             self._write_sp()
-            self._format_child() # '{'
-            self._write_nl()
-            self._format_child(indent=True) # enum_body
-            self._format_child() # '}'
+            self._format_curly_enum_body()
 
         elif self._get_child_token() == 'function':
             self._format_child_range(2) # 'function' <func_params>
@@ -545,11 +739,23 @@ class TypeSpecFormatter(Formatter):
 
 class EnumBodyFormatter(Formatter):
     def format(self):
-        while self._get_child():
-            self._format_child() # enum_body_elem
-            if self._get_child():
-                self._format_child(hints=Hint.NO_LB_BEFORE) # ',' (optional at the end of the list)
-            self._write_nl()
+        if Hint.COMPLEX_BLOCK in self.hints:
+            # Treat this as a "complex": break every value onto a new line.
+            while self._get_child():
+                self._format_child() # enum_body_elem
+                # ',' is optional at the end of the list:
+                if self._get_child():
+                    self._format_child(hints=Hint.NO_LB_BEFORE)
+                self._write_nl()
+        else:
+            # Keep on a single line. We may still linewrap later.
+            while self._get_child():
+                self._format_child() # enum_body_elem
+                # ',' is optional at the end of the list:
+                if self._get_child():
+                    self._format_child(hints=Hint.NO_LB_BEFORE)
+                if self._get_child():
+                    self._write_sp()
 
 
 class FuncDeclFormatter(Formatter):
@@ -565,6 +771,7 @@ class FuncDeclFormatter(Formatter):
         self._format_child() # <func_body>
         self._write_nl()
 
+
 class FuncHdrFormatter(Formatter):
     def format(self):
         self._format_child() # <func>, <hook>, or <event>
@@ -598,13 +805,8 @@ class FuncParamsFormatter(Formatter):
 
 class FuncBodyFormatter(Formatter):
     def format(self):
-        self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
-        if self._get_child_name() == 'stmt_list':
-            self._write_nl()
-            self._format_child(indent=True) # <stmt_list>
-        else:
-            self._write_sp()
-        self._format_child() # '}'
+        self._write_nl()
+        self._format_curly_statement_list()
 
 
 class FormalArgsFormatter(Formatter):
@@ -627,6 +829,16 @@ class FormalArgFormatter(Formatter):
             self._format_child() # <attr_list>
 
 
+class IndexSliceFormatter(Formatter):
+    def format(self):
+        self._format_child(hints=Hint.NO_LB_BEFORE) # '['
+        while self._get_child_token() != ']':
+            self._format_child()
+            if self._get_child_token() != ']':
+                self._write_sp()
+        self._format_child(hints=Hint.NO_LB_BEFORE) # ']'
+
+
 class CaptureListFormatter(Formatter):
     def format(self):
         self._format_child(hints=Hint.NO_LB_BEFORE) # '['
@@ -636,35 +848,17 @@ class CaptureListFormatter(Formatter):
                 self._format_child(hints=Hint.NO_LB_BEFORE) # ','
                 self._write_sp()
         self._format_child(hints=Hint.NO_LB_BEFORE) # ']'
-        self._write_sp()
 
 
 class StmtFormatter(TypedInitializerFormatter):
-    def _child_is_curly_stmt(self):
-        """Looks ahead to see if the upcoming statement is { ... }.
-        This decides surrounding whitespace in some situations below.
+    def _format_stmt_block(self):
+        """Helper for formatting a block of statements.
+
+        This may either be an { ... } block or a single-line statement.
         """
-        # This checks a property of the child's children: to trigger, the child
-        # is an if- or else-block, and 'if' is the first child token in that
-        return self._get_child().has_property(lambda n: n.children[0].token() == '{')
-
-    def _write_sp_or_nl(self, do_sp):
-        """Writes separator based on sp_or_nl.
-
-        This guides whitespace depending on whether if et al. have a {}-block.
-        """
-        if do_sp:
-            self._write_sp()
-        else:
-            self._write_nl()
-
-    def _format_block(self):
-        """Helper for formatting a statement that may be an { ... } block."""
-        curly = self._child_is_curly_stmt()
-        self._write_sp_or_nl(curly)
-        self._format_child(indent=not curly) # <stmt>
-        if curly:
-            self._write_nl()
+        self._write_nl()
+        self._format_child(indent=True) # <stmt>
+        self._write_nl()
 
     def _format_when(self):
         self._format_child() # 'when'
@@ -677,26 +871,15 @@ class StmtFormatter(TypedInitializerFormatter):
         self._format_child() # <expr>
         self._write_sp()
         self._format_child(hints=Hint.NO_LB_BEFORE) # ')'
-
-        curly = self._child_is_curly_stmt()
-        self._write_sp_or_nl(curly)
-        self._format_child(indent=not curly) # <stmt>
+        self._format_stmt_block()
 
         if self._get_child_token() == 'timeout':
-            if curly:
-                self._write_sp()
             self._format_child() # 'timeout'
             self._write_sp()
             self._format_child() # <expr>
-            self._write_sp()
-            self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
             self._write_nl()
-            if self._get_child_name() == 'stmt_list':
-                self._format_child(indent=True) # <stmt_list>
-            self._format_child() # '}'
+            self._format_curly_statement_list() # '{' <stmt_list> '}'
             self._write_nl()
-        elif curly:
-            self._write_nl() # Finish the when's curly block.
 
     def format(self):
         # Statements aren't currently broken down into more specific symbol
@@ -704,13 +887,9 @@ class StmtFormatter(TypedInitializerFormatter):
         start_name, start_token = self._get_child_name(), self._get_child_token()
 
         if start_token == '{':
-            self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
-            if self._get_child_name() == 'stmt_list':
-                self._write_nl()
-                self._format_child(indent=True)
-            else:
-                self._write_sp()
-            self._format_child() # '}'
+            # We don't have to do anything re. Whitesmith here: if this needs
+            # to be indented, the caller has already ensured so via indent=True.
+            self._format_curly_statement_list(indent=False) # '{' <stmt_list> '}'
 
         elif start_token in ['print', 'event']:
             self._format_child() # 'print'/'event'
@@ -735,52 +914,49 @@ class StmtFormatter(TypedInitializerFormatter):
             #
             # or
             #
-            #   if ( foo ) {
+            #   if ( foo )
+            #           {
             #           bar();
-            #   } ...
+            #           }
+            #   ...
             #
             # We need to establish whether the subsequent statement is a
-            # {}-block, because if it's not we write a newline and need to
-            # indent, because {}-blocks take care of indentation as another
-            # statement type (higher up in this function).
+            # {}-block: if it's not, we write a newline and need to indent,
+            # because {}-blocks take care of indentation as another statement
+            # type (higher up in this function).
 
-            curly = self._child_is_curly_stmt()
-            self._write_sp_or_nl(curly)
-            self._format_child(indent=not curly) # <stmt>
+            self._format_stmt_block()
 
             # An else-block also requires special treatment
             if self._get_child_token() == 'else':
-                if curly:
-                    self._write_sp()
                 self._format_child() # 'else'
 
                 # Special treatment of "else if": we keep those on the same
                 # line, since otherwise, a switch-case-like cascade of if-else
                 # would get progressively more indented.
-                if self._get_child().has_property(lambda n: n.children[0].token() == 'if'):
+                if self._get_child().has_property(lambda n: n.nonerr_children[0].token() == 'if'):
                     self._write_sp()
                     self._format_child() # <stmt>
                 else:
-                    curly = self._child_is_curly_stmt()
-                    self._write_sp_or_nl(curly)
-                    self._format_child(indent=not curly) # <stmt>
-                    if curly:
-                        self._write_nl()
-            elif curly:
-                self._write_nl() # Finish the if's curly block.
+                    self._format_stmt_block()
 
         elif start_token == 'switch':
             self._format_child() # 'switch'
             self._write_sp()
             self._format_child() # <expr>
-            self._write_sp()
-            self._format_child(hints=Hint.NO_LB_BEFORE) # '{'
-            if self._get_child_name() == 'case_list':
-                self._write_nl()
-                self._format_child(indent=True) # <case_list>
-            else:
+            self._write_nl()
+            self._format_child(indent=True) # '{'
+            # Shorten braces to "{ }" if there is at most whitespace between them.
+            if (self._get_child_token() == '}' and
+                self._get_child().has_only_whitespace_before()):
                 self._write_sp()
-            self._format_child() # '}'
+                self._format_child() # '}'
+            else:
+                if self._get_child_name() == 'case_list':
+                    self._write_nl()
+                    self._format_child(indent=True) # <case_list>
+                self._write_nl()
+                self._format_child(indent=True) # '}'
             self._write_nl()
 
         elif start_token == 'for':
@@ -809,7 +985,7 @@ class StmtFormatter(TypedInitializerFormatter):
             self._format_child() # <expr>
             self._write_sp()
             self._format_child(hints=Hint.NO_LB_BEFORE) # ')'
-            self._format_block() # <stmt>
+            self._format_stmt_block() # <stmt>
 
         elif start_token == 'while':
             self._format_child() # 'while'
@@ -819,7 +995,7 @@ class StmtFormatter(TypedInitializerFormatter):
             self._format_child() # <expr>
             self._write_sp()
             self._format_child(hints=Hint.NO_LB_BEFORE) # ')'
-            self._format_block() # <stmt>
+            self._format_stmt_block() # <stmt>
 
         elif start_token in ['next', 'break', 'fallthrough']:
             self._format_child_range(2) # loop control statement, ';'
@@ -873,17 +1049,24 @@ class StmtFormatter(TypedInitializerFormatter):
             self._write_nl()
 
         elif start_token == ';':
-            self._format_child() # ';'
+            self._format_child(hints=Hint.NO_LB_BEFORE) # ';'
             self._write_nl()
 
 
-class ExprListFormatter(Formatter):
+class ExprListFormatter(Formatter, ComplexSequenceFormatterMixin):
     def format(self):
-        while self._get_child_name() == 'expr':
-            self._format_child() # <expr>
-            if self._get_child():
-                self._format_child(hints=Hint.NO_LB_BEFORE) # ','
-                self._write_sp()
+        if self.is_complex():
+            while self._get_child_name() == 'expr':
+                self._format_child(indent=True) # <expr>
+                if self._get_child():
+                    self._format_child(hints=Hint.NO_LB_BEFORE) # ','
+                    self._write_nl()
+        else:
+            while self._get_child_name() == 'expr':
+                self._format_child() # <expr>
+                if self._get_child():
+                    self._format_child(hints=Hint.NO_LB_BEFORE) # ','
+                    self._write_sp()
 
 
 class CaseListFormatter(Formatter):
@@ -932,13 +1115,28 @@ class ExprFormatter(SpaceSeparatedFormatter):
 
     def _is_binary_boolean(self):
         """Predicate, returns true if this an || or && expression."""
-        return (len(self.node.children) == 3 and
+        return (len(self.node.nonerr_children) == 3 and
                 self._get_child_token(offset=1, absolute=True) in ('||', '&&'))
 
-    def _is_binary_addition(self):
-        """Predicate, returns true if this an <expr> + <expr> expression."""
-        return (len(self.node.children) == 3 and
-                self._get_child_type(offset=1, absolute=True) == '+')
+    def _is_string_concat(self):
+        """Predicate, returns true if this a <string> + <string> expression."""
+        def is_constant_expr(node):
+            return (node.name() == 'expr' and
+                    len(node.nonerr_children) == 1 and
+                    node.nonerr_children[0].name() == 'constant' and
+                    len(node.nonerr_children[0].nonerr_children) == 1 and
+                    node.nonerr_children[0].nonerr_children[0].name() == 'string')
+
+        def is_concat_expr(node):
+            return (node.name() == 'expr' and
+                    len(node.nonerr_children) == 3 and
+                    (is_constant_expr(node.nonerr_children[0]) or
+                     is_concat_expr(node.nonerr_children[0])) and
+                    node.nonerr_children[1].token() == '+' and
+                    (is_constant_expr(node.nonerr_children[2]) or
+                     is_concat_expr(node.nonerr_children[2])))
+
+        return is_concat_expr(self.node)
 
     def _is_expr_chain_of(self, formatter_predicate):
         """Predicate, returns true if the given predicate is true for all
@@ -1003,10 +1201,10 @@ class ExprFormatter(SpaceSeparatedFormatter):
             self._format_child(hints=Hint.NO_LB_BEFORE) # ']
 
         elif ct1 == '$' and ct3 == '=':
-            self._format_child_range(4, first_hints=Hint.GOOD_AFTER_LB) # '$'<id> = <expr>
+            self._format_child_range(4) # '$'<id> = <expr>
 
         elif ct1 == '$': # The function version, with possible capture
-            self._format_child_range(2, first_hints=Hint.GOOD_AFTER_LB) # '$'<id>
+            self._format_child_range(2) # '$'<id>
             self._write_sp()
             self._format_child(hints=Hint.NO_LB_BEFORE | Hint.NO_LB_AFTER) # <begin_lambda>
             self._write_sp()
@@ -1068,17 +1266,12 @@ class ExprFormatter(SpaceSeparatedFormatter):
             self._write_sp()
             self._format_child() # <expr>
 
-        elif self._is_binary_addition():
-            # Same approach, but for additions. This helps OutputStream nicely
-            # align long strings broken into substrings concatenated by "+".
-            hints = None
-
-            if self._is_expr_chain_of(ExprFormatter._is_binary_addition):
-                hints = Hint.GOOD_AFTER_LB
-
+        elif self._is_string_concat():
+            # This helps OutputStream nicely align long strings broken into
+            # substrings concatenated by "+".
             self._format_child() # <expr>
             self._write_sp()
-            self._format_child(hints=hints) # '+'
+            self._format_child(hints=Hint.GOOD_AFTER_LB) # '+'
             self._write_sp()
             self._format_child() # <expr>
 
@@ -1131,7 +1324,7 @@ class CommentFormatter(Formatter):
     """Base class for any kind of comment."""
     def __init__(self, script, node, ostream, indent=0, hints=None):
         super().__init__(script, node, ostream, indent, hints)
-        self.hints |= Hint.ZERO_WIDTH # Commens never count toward line length
+        self.hints |= Hint.ZERO_WIDTH # Comments never count toward line length
 
 
 class MinorCommentFormatter(CommentFormatter):
@@ -1211,6 +1404,7 @@ Formatter.register('event', FuncHdrVariantFormatter)
 Formatter.register('capture', SpaceSeparatedFormatter)
 Formatter.register('attr_list', SpaceSeparatedFormatter)
 Formatter.register('interval', SpaceSeparatedFormatter)
+Formatter.register('enum_body_elem', SpaceSeparatedFormatter)
 
 Formatter.register('zeekygen_head_comment', ZeekygenCommentFormatter)
 Formatter.register('zeekygen_next_comment', ZeekygenCommentFormatter)
