@@ -16,11 +16,13 @@ class Output:
 
 class OutputStream:
     """An indenting, column-aware, line-buffered, line-wrapping,
-    trailing-whitespace-stripping output stream wrapper.
+    trailing-whitespace-stripping output stream wrapper. When used as a context
+    manager, it also ensures that exiting context always finishes output with a
+    newline.
     """
     MAX_LINE_LEN = 80 # Column at which we consider wrapping.
     MIN_LINE_ITEMS = 5 # Required items on a line to consider wrapping.
-    MIN_LINE_EXCESS = 10 # Minimum characters that a line needs to be too long.
+    MIN_LINE_EXCESS = 5 # Minimum characters that a line needs to be too long.
     TAB_SIZE = 8 # How many visible characters we chalk up for a tab.
     SPACE_INDENT = 4 # When wrapping, add this many spaces onto tab-indentation.
 
@@ -30,9 +32,14 @@ class OutputStream:
         self._col = 0 # 0-based column the next character goes into.
         self._tab_indent = 0 # Number of tabs indented in current line
 
-        # Series of Output objects that makes up a formatted but unbroken line.
+        # Series of Output objects that makes up a formatted but un-wrapped line.
         self._linebuffer = []
 
+        # Series of byte sequences actually written out, post line-wrap. This is
+        # line-buffered, to allow removal of trailing whitespace. Details in
+        # self._write().
+        self._writebuffer = []
+
         # Whether we'll consider linebreaks at all. When False, long lines will
         # never wrap. When True, linebreaks will generally happen, but
         # formatters may pause them temporarily via the _use_linebreaks flag
@@ -48,6 +55,13 @@ class OutputStream:
         # line; they just add a few spaces.
         self._use_space_align = False
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
+        self._flush_line()
+        self._flush_writes()
+
     def use_linebreaks(self, enable):
         self._use_linebreaks = enable
 
@@ -57,25 +71,24 @@ class OutputStream:
     def use_space_align(self, enable):
         self._use_space_align = enable
 
-    def write(self, data, formatter):
+    def write(self, data, formatter, raw=False):
+        if raw:
+            self._flush_line()
+            self._write(data)
+            # Sync column count as per last line content in raw data:
+            self._col = len(data.split(Formatter.NL)[-1])
+            return
+
         # For troubleshooting received hinting
         # print_error('XXX "%s" %s' % (data, formatter.hints))
 
+        # In case the data spans multiple lines, break up the lines now.
+        # Potential line-wrapping is applied when we flush individual lines.
         for chunk in data.splitlines(keepends=True):
-            if chunk.endswith(Formatter.NL):
-                # Remove any trailing whitespace
-                chunk = chunk.rstrip() + Formatter.NL
-
             self._col += len(chunk)
-
-            if self._enable_linebreaks and self._use_linebreaks:
-                self._linebuffer.append(Output(chunk, formatter))
-                if chunk.endswith(Formatter.NL):
-                    self._flush_line()
-            else:
-                self._write(chunk)
-                if chunk.endswith(Formatter.NL):
-                    self._col = 0
+            self._linebuffer.append(Output(chunk, formatter))
+            if chunk.endswith(Formatter.NL):
+                self._flush_line()
 
     def write_tab_indent(self, formatter):
         if self._use_tab_indent:
@@ -90,13 +103,23 @@ class OutputStream:
         return self._col
 
     def _flush_line(self):
-        """Helper that flushes out the built-up line buffer.
+        """Flushes out the line buffer, stripping trailing whitespace.
 
-        This iterates over the Output objects in self._linebuffer, deciding
-        whether to write them out right away or in to-be-done batches, possibly
+        Without linewrapping, this simply flushes the line buffer. With
+        linewrapping this iterates over the line buffer, deciding whether to
+        write out data chunks right away or in to-be-done batches, possibly
         after newlines, depending on line-breaking hints present in the
         formatter objects linked from the Output instances.
         """
+        # Without linebreaking active, just flush the buffer.
+        if not self._enable_linebreaks or not self._use_linebreaks:
+            for out in self._linebuffer:
+                self._write(out.data)
+
+            self._linebuffer = []
+            self._col = 0
+            return
+
         col_flushed = 0 # Column up to which we've currently written a line
         tbd = [] # Outputs to be done
         tbd_len = 0 # Length of the to-be-done output (in characters)
@@ -130,9 +153,9 @@ class OutputStream:
                 line_items += 1
 
         # It is logistically more difficult to honor NO_LB_BEFORE as it arises,
-        # because we need to cover all cases where such a "look-ahead" prevents
-        # breaking. To simplify, we reverse-iterate over the line's tokens
-        # and tuck NO_LB_AFTER onto the that precede NO_LB_BEFORE ones.
+        # because we need to "look-ahead" to prevent breaking. To simplify,
+        # reverse-iterate over the line's tokens and tuck NO_LB_AFTER onto
+        # tokens that precede NO_LB_BEFORE.
         needs_no_lb_after = False
         for out in self._linebuffer[::-1]:
             if out.data.strip() and needs_no_lb_after:
@@ -149,52 +172,68 @@ class OutputStream:
             if Hint.ZERO_WIDTH not in out.formatter.hints:
                 tbd_len += len(out.data)
 
-            # Don't write mid-line whitespace right away: if content follows
-            # it that gets wrapped, we'd produce trailing whitespace. We instead
-            # push such whitespace onto the next line, where write_linebreak()
-            # suppresses it when needed.
+            # Don't make line-wrapping decisions based on whitespace:
             if not out.data.strip():
                 continue
 
+            # We name the various conditions going into the linebreak decision,
+            # so we can report them for troubleshooting and act on them below
+
+            # If the line is too long and this chunk says it best follows a
+            # break, then break now. This helps align e.g. multi-part boolean
+            # conditionals. This needs to take precedence over NO_LB_AFTER,
+            # see next condition.
+            cnd_good_after_lb = (Hint.GOOD_AFTER_LB in out.formatter.hints and
+                                 self._col > self.MAX_LINE_LEN)
+
+            # If the caller requested no line break, abide.
+            cnd_no_lb_after = Hint.NO_LB_AFTER in out.formatter.hints
+
+            # Similarly, if we git GOOD_AFTER_LB earlier, abide.
+            cnd_no_break_hints = not using_break_hints
+
+            # We need to exceed the MAX_LINE_LEN limit with what's pending.
+            cnd_line_too_long = col_flushed + tbd_len > self.MAX_LINE_LEN
+
+            # The pending length must be "worth it". That is, don't break if the
+            # TBD len is just a little bit over. But do so if we're just too
+            # long overall now.
+            cnd_enough_excess = (tbd_len >= self.MIN_LINE_EXCESS or
+                                 col_flushed > self.MAX_LINE_LEN + self.MIN_LINE_EXCESS)
+
+            # If there are only very few items on the line to begin with, don't
+            # bother: breaking these also looks messy. This often covers the
+            # case of a line consisting mostly of a long string.
+            cnd_enough_line_items = line_items >= self.MIN_LINE_ITEMS
+
+            # If the TBD items would immediately exceed the line limit again
+            # after wrapping, don't bother. This covers the case of e.g. very
+            # long strings looking silly when alone on a new line. (This doesn't
+            # interfere with bit-by-bit repeated linebreaks of a very long line
+            # -- that still happens when we build up the next TBD batch.)
+            cnd_no_addl_wrap =  self._tab_indent * self.TAB_SIZE + tbd_len < self.MAX_LINE_LEN
+
+            # Helpful for tracing linebreak decision-making:
+            # print_error('XXX gal:%d nla:%d nbh:%d tl:%d ex:%d ei:%d naw:%d | %s %s %s' % (
+            #    cnd_good_after_lb, cnd_no_lb_after, cnd_no_break_hints,
+            #    cnd_line_too_long, cnd_enough_excess, cnd_enough_line_items,
+            #    cnd_no_addl_wrap, out.data, col_flushed, tbd_len))
+
             # If the line is too long and this chunk says it best follows a
             # break, then break now. This helps align e.g. multi-part boolean
             # conditionals. This needs to take precedence over NO_LB_AFTER.
-            elif Hint.GOOD_AFTER_LB in out.formatter.hints and self._col > self.MAX_LINE_LEN:
+            if cnd_good_after_lb:
                 write_linebreak()
                 using_break_hints = True
 
             # Honor hinted linebreak suppression around this chunk.
-            elif Hint.NO_LB_AFTER in out.formatter.hints:
+            elif cnd_no_lb_after:
                 continue
 
-            # Finally actually linebreak as needed, possibly repeatedly:
-            #
-            # - If we used the GOOD_AFTER_LB hint above, rely exclusively on it
-            #   for future breaks, because a mix tends to look messy. So don't
-            #   break here in that case.
-            #
-            # - We need to exceed the MAX_LINE_LEN column limit by writing
-            #   what's pending.
-            #
-            # - The pending length must be "worth it". That is, don't break
-            #   if the TBD len is just a little bit.
-            #
-            # - If there are only very few items on the line to begin with,
-            #   don't bother: breaking these also looks messy. This often covers
-            #   the case of a line consisting mostly of a long string.
-            #
-            # - If the TBD items would immediately exceed the line limit again
-            #   after wrapping, don't bother. This covers the case of e.g. very
-            #   long strings looking silly when alone on a new line. (This
-            #   doesn't interfere with bit-by-bit repeated linebreaks of a very
-            #   long line -- that still happens when we build up the next TBD
-            #   batch.)
-            #
-            elif (not using_break_hints
-                  and col_flushed + tbd_len > self.MAX_LINE_LEN
-                  and tbd_len >= self.MIN_LINE_EXCESS
-                  and line_items >= self.MIN_LINE_ITEMS
-                  and self._tab_indent * self.TAB_SIZE + tbd_len < self.MAX_LINE_LEN):
+            # Finally actually linebreak as needed:
+            elif (cnd_no_break_hints and cnd_line_too_long and
+                  cnd_enough_excess and cnd_enough_line_items and
+                  cnd_no_addl_wrap):
                 write_linebreak()
 
             flush_tbd()
@@ -205,14 +244,30 @@ class OutputStream:
         self._linebuffer = []
         self._col = 0
 
+    def _flush_writes(self):
+        if self._writebuffer and not self._writebuffer[-1].endswith(Formatter.NL):
+            self._write(Formatter.NL)
+
     def _write(self, data):
+        self._writebuffer.append(data)
+
+        # Data can only have newlines at the end, since we already split any
+        # mid-data newlines earlier. So check for trailing newlines, and if
+        # found, strip trailing whitespace and flush.
+        if not data.endswith(Formatter.NL):
+            return
+
+        output = b''.join(self._writebuffer)
+        output = output.rstrip() + Formatter.NL
+        self._writebuffer = []
+
         try:
             if self._ostream == sys.stdout:
                 # Clunky: must write string here, not bytes. We could
                 # use _ostream.buffer -- not sure how portable that is.
-                self._ostream.write(data.decode('UTF-8'))
+                self._ostream.write(output.decode('UTF-8'))
             else:
-                self._ostream.write(data)
+                self._ostream.write(output)
         except BrokenPipeError:
             #  https://docs.python.org/3/library/signal.html#note-on-sigpipe:
             devnull = os.open(os.devnull, os.O_WRONLY)
